import json
import logging
from copy import deepcopy
from urllib3 import Retry
from binascii import unhexlify
try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

import requests
import datetime as dt
from dateutil import parser
from future.utils import raise_from
from requests.adapters import HTTPAdapter
from urllib3.exceptions import InsecureRequestWarning

from polyswarm_api import settings, exceptions

logger = logging.getLogger(__name__)


class PolyswarmSession(requests.Session):
    def __init__(self, key, retries, user_agent=settings.DEFAULT_USER_AGENT, verify=True, **kwargs):
        super(PolyswarmSession, self).__init__(**kwargs)
        logger.debug('Creating PolyswarmHTTP instance')
        self.requests_retry_session(retries=retries)

        if not verify:
            logger.warn('Disabling TLS verification for this session.')
            requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

        self.verify = verify

        if key:
            self.set_auth(key)

        if user_agent:
            self.set_user_agent(user_agent)

    def requests_retry_session(self, retries=settings.DEFAULT_RETRIES, backoff_factor=settings.DEFAULT_BACKOFF,
                               status_forcelist=settings.DEFAULT_RETRY_CODES):
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )
        adapter = HTTPAdapter(max_retries=retry)
        self.mount('http://', adapter)
        self.mount('https://', adapter)

    def set_auth(self, key):
        if key:
            self.headers.update({'Authorization': key})
        else:
            self.headers.pop('Authorization', None)

    def set_user_agent(self, ua):
        if ua:
            self.headers.update({'User-Agent': ua})
        else:
            self.headers.pop('User-Agent', None)


class RequestParamsEncoder(json.JSONEncoder):
    def default(self, obj):
        try:
            return json.JSONEncoder.default(self, obj)
        except Exception:
            return str(obj)


class PolyswarmRequest(object):
    """This class holds a requests-compatible dictionary and extra information we need to parse the response."""
    def __init__(self, api_instance, request_parameters, key=None, result_parser=None, **kwargs):
        logger.debug('Creating PolyswarmRequest instance.\nRequest parameters: %s\nResult parser: %s',
                     request_parameters, result_parser.__name__ if result_parser else 'No result parser')
        self.api_instance = api_instance
        # we should not access the api_instance session directly, but provide as a
        # parameter in the constructor, but this will do for the moment
        self.session = self.api_instance.session or PolyswarmSession(key, retries=settings.DEFAULT_RETRIES)
        self.timeout = self.api_instance.timeout or settings.DEFAULT_HTTP_TIMEOUT
        self.request_parameters = request_parameters
        self.result_parser = result_parser
        self.raw_result = None
        self.status_code = None
        self.status = None
        self.errors = None
        self._result = None

        self._paginated = False
        self.total = None
        self.limit = None
        self.offset = None
        self.order_by = None
        self.direction = None
        self.has_more = None

        self.parser_kwargs = kwargs

    def result(self):
        if self._paginated:
            return self.consume_results()
        else:
            return self._result

    def execute(self):
        logger.debug('Executing request.')
        self.request_parameters.setdefault('timeout', self.timeout)
        if self.result_parser and not issubclass(self.result_parser, BaseJsonResource):
            self.request_parameters.setdefault('stream', True)
        self.raw_result = self.session.request(**self.request_parameters)
        logger.debug('Request returned code %s', self.raw_result.status_code)
        self.parse_result(self.raw_result)
        return self

    def _bad_status_message(self):
        request_parameters = json.dumps(self.request_parameters, indent=4, sort_keys=True, cls=RequestParamsEncoder)
        message = "Error when running the request:\n{}\n" \
                  "Return code: {}\n" \
                  "Message: {}".format(request_parameters,
                                       self.status_code,
                                       self._result)
        if self.errors:
            message = '{}\nErrors:\n{}'.format(message, '\n'.join(str(error) for error in self.errors))
        return message

    def _extract_json_body(self, result):
        self.json = result.json()
        self._result = self.json.get('result')
        self.status = self.json.get('status')
        self.errors = self.json.get('errors')

    def parse_result(self, result):
        self.status_code = result.status_code
        if self.request_parameters['method'] == 'HEAD':
            logger.debug('HEAD method does not return results, setting it to the status code.')
            self._result = self.status_code
        if not self.result_parser:
            logger.debug('Result parser is not defined, skipping parsing results.')
            return
        logger.debug('Parsing request results.')
        try:
            if self.status_code // 100 != 2:
                self._extract_json_body(result)
                if self.status_code == 429:
                    message = '{} This may mean you need to purchase a ' \
                              'larger package, or that you have exceeded ' \
                              'rate limits. If you continue to have issues, ' \
                              'please contact us at info@polyswarm.io.'.format(self._result)
                    raise exceptions.UsageLimitsExceededException(self, message)
                elif self.status_code == 404:
                    raise exceptions.NotFoundException(self, self._result)
                else:
                    raise exceptions.RequestException(self, self._bad_status_message())
            elif self.status_code == 204:
                raise exceptions.NoResultsException(self, 'The request returned no results.')
            elif issubclass(self.result_parser, BaseJsonResource):
                self._extract_json_body(result)
                if self.request_parameters['method'] == 'GET' and 'has_more' in self.json:
                    # has_more will always be present, being either False or True
                    self._paginated = True
                self.total = self.json.get('total')
                self.limit = self.json.get('limit')
                self.offset = self.json.get('offset')
                self.order_by = self.json.get('order_by')
                self.direction = self.json.get('direction')
                self.has_more = self.json.get('has_more')
                if 'result' in self.json:
                    result = self.json['result']
                elif 'results' in self.json:
                    result = self.json['results']
                else:
                    raise exceptions.RequestException(
                        self,
                        'The response standard must contain either the "result" or "results" key.'
                    )
                if isinstance(result, list):
                    self._result = self.result_parser.parse_result_list(self.api_instance, result, **self.parser_kwargs)
                else:
                    self._result = self.result_parser.parse_result(self.api_instance, result, **self.parser_kwargs)
            else:
                self._result = self.result_parser.parse_result(self.api_instance, result, **self.parser_kwargs)
        except JSONDecodeError as e:
            if self.status_code == 404:
                raise raise_from(exceptions.NotFoundException(self, 'The requested endpoint does not exist.'), e)
            else:
                err_msg = 'Server returned non-JSON response [{}]: {}'.format(self.status_code, result)
                raise raise_from(exceptions.RequestException(self, err_msg), e)

    def __iter__(self):
        return self.consume_results()

    def consume_results(self):
        # StopIteration is deprecated
        # As per https://www.python.org/dev/peps/pep-0479/
        # We simply return upon termination condition
        request = self
        while True:
            # consume items items from list if iterable
            # of yield the single result if not
            try:
                for result in request._result:
                    yield result
            except TypeError:
                yield request._result
                # if the result is not a list, there is not next page
                return

            # if the server indicates that there are no more results, return
            if not request.has_more:
                return
            # try to get the next page and execute the request
            request = request.next_page()

    def next_page(self):
        new_parameters = deepcopy(self.request_parameters)
        params = new_parameters.setdefault('params', {})
        if isinstance(params, dict):
            params['offset'] = self.offset
            params['limit'] = self.limit
        else:
            params = [p for p in params if p[0] != 'offset' and p[0] != 'limit']
            params.extend([('offset', self.offset), ('limit', self.limit)])
            new_parameters['params'] = params
        return PolyswarmRequest(
            self.api_instance,
            new_parameters,
            result_parser=self.result_parser,
        ).execute()


class BaseResource(object):
    def __init__(self, content, *args, **kwargs):
        # hack to behave as in python 3, signature should be
        # __init__(self, content, *args, api=None, **kwargs)
        api = kwargs.pop('api', None)
        super(BaseResource, self).__init__(*args, **kwargs)
        self.api = api
        self._content = content

    @classmethod
    def parse_result(cls, api, content, **kwargs):
        logger.debug('Parsing resource %s', cls.__name__)
        return cls(content, api=api, **kwargs)


class BaseJsonResource(BaseResource):
    RESOURCE_ENDPOINT = None
    RESOURCE_ID_KEY = 'id'

    def __init__(self, content, *args, **kwargs):
        super(BaseJsonResource, self).__init__(content, *args, **kwargs)
        self.json = content

    def __int__(self):
        id_ = getattr(self, 'id', None)
        if id_ is None:
            raise TypeError('Resource {} does not have an id and can not be cast to int'.format(type(self).__name__))
        return int(id_)

    def _get(self, path, default=None, content=None):
        """
        Helper for rendering attributes of child objects in the json that might be None.
        Returns the default value if any item in the path is not present.
        """
        previous_attribute = 'resource_json'
        obj = content or self.json
        try:
            for attribute in path.split('.'):
                if obj is None:
                    raise KeyError('{} is None, can not resolve full path'.format(previous_attribute))
                if attribute.endswith(']'):
                    # handling the list case, e.g.: "root.list_attr[2]"
                    attribute, _, index = attribute.rpartition('[')
                    index = int(index.rstrip(']'))
                    obj = obj[attribute]
                    if obj is None:
                        raise KeyError('{} is None, but is it supposed to be a list'.format(attribute))
                    elif not isinstance(obj, list):
                        raise ValueError('Can not access index for {}, it is not a list.'.format(attribute))
                    else:
                        obj = obj[index]
                else:
                    obj = obj[attribute]
                previous_attribute = attribute
            return obj
        except (KeyError, IndexError) as e:
            logger.debug('Returning default value: %s', e)
            return default

    @classmethod
    def parse_result_list(cls, api_instance, json_data, **kwargs):
        return [cls.parse_result(api_instance, entry, **kwargs) for entry in json_data]

    @classmethod
    def _endpoint(cls, api, **kwargs):
        if cls.RESOURCE_ENDPOINT is None:
            raise exceptions.InvalidValueException('RESOURCE_ENDPOINT is not configured for this resource.')
        return '{api.uri}{endpoint}'.format(api=api, endpoint=cls.RESOURCE_ENDPOINT, **kwargs)

    @classmethod
    def _list_endpoint(cls, api, **kwargs):
        return cls._endpoint(api, **kwargs) + '/list'

    @classmethod
    def _create_endpoint(cls, api, **kwargs):
        return cls._endpoint(api, **kwargs)

    @classmethod
    def _get_endpoint(cls, api, **kwargs):
        return cls._endpoint(api, **kwargs)

    @classmethod
    def _head_endpoint(cls, api, **kwargs):
        return cls._endpoint(api, **kwargs)

    @classmethod
    def _update_endpoint(cls, api, **kwargs):
        return cls._endpoint(api, **kwargs)

    @classmethod
    def _delete_endpoint(cls, api, **kwargs):
        return cls._endpoint(api, **kwargs)

    @classmethod
    def _params(cls, method, *param_keys, **kwargs):
        params = {}
        json_params = {}
        for k, v in kwargs.items():
            if v is not None:
                # try to parse "*_id" stuff as integer
                if k.endswith('_id'):
                    try:
                        parsed_value = str(int(v))
                    except Exception:
                        # fallback to string
                        parsed_value = str(v)
                elif isinstance(v, bool):
                    parsed_value = int(v)
                else:
                    parsed_value = v
                if method == 'POST':
                    json_params[k] = parsed_value
                elif method == 'GET' or k in param_keys:
                    params[k] = parsed_value
                else:
                    json_params[k] = parsed_value

        params = params if params else None
        json_params = json_params if json_params else None
        return params, json_params

    @classmethod
    def _list_params(cls, **kwargs):
        return cls._params('GET', cls.RESOURCE_ID_KEY, **kwargs)

    @classmethod
    def _create_params(cls, **kwargs):
        return cls._params('POST', cls.RESOURCE_ID_KEY, **kwargs)

    @classmethod
    def _get_params(cls, **kwargs):
        return cls._params('GET', cls.RESOURCE_ID_KEY, **kwargs)

    @classmethod
    def _head_params(cls, **kwargs):
        return cls._params('HEAD', cls.RESOURCE_ID_KEY, **kwargs)

    @classmethod
    def _update_params(cls, **kwargs):
        return cls._params('PUT', cls.RESOURCE_ID_KEY, **kwargs)

    @classmethod
    def _delete_params(cls, **kwargs):
        return cls._params('DELETE', cls.RESOURCE_ID_KEY, **kwargs)

    @classmethod
    def _list_headers(cls, api):
        return None

    @classmethod
    def _create_headers(cls, api):
        return None

    @classmethod
    def _get_headers(cls, api):
        return None

    @classmethod
    def _head_headers(cls, api):
        return None

    @classmethod
    def _update_headers(cls, api):
        return None

    @classmethod
    def _delete_headers(cls, api):
        return None

    @classmethod
    def _build_request(cls, api, method, url, headers, params, json_params):
        request_params = {'method': method, 'url': url}
        if params:
            request_params['params'] = params
        if json_params:
            request_params['json'] = json_params
        if headers:
            request_params['headers'] = headers
        return PolyswarmRequest(api, request_params, result_parser=cls)

    @classmethod
    def create(cls, api, **kwargs):
        return cls._build_request(api, 'POST', cls._create_endpoint(api, **kwargs),
                                  cls._create_headers(api), *cls._create_params(**kwargs)).execute()

    @classmethod
    def get(cls, api, **kwargs):
        return cls._build_request(api, 'GET', cls._get_endpoint(api, **kwargs),
                                  cls._get_headers(api), *cls._get_params(**kwargs)).execute()

    @classmethod
    def head(cls, api, **kwargs):
        return cls._build_request(api, 'HEAD', cls._head_endpoint(api, **kwargs),
                                  cls._head_headers(api), *cls._head_params(**kwargs)).execute()

    @classmethod
    def update(cls, api, **kwargs):
        return cls._build_request(api, 'PUT', cls._update_endpoint(api, **kwargs),
                                  cls._update_headers(api), *cls._update_params(**kwargs)).execute()

    @classmethod
    def delete(cls, api, **kwargs):
        return cls._build_request(api, 'DELETE', cls._delete_endpoint(api, **kwargs),
                                  cls._delete_headers(api), *cls._delete_params(**kwargs)).execute()

    @classmethod
    def list(cls, api, **kwargs):
        return cls._build_request(api, 'GET', cls._list_endpoint(api, **kwargs),
                                  cls._list_headers(api), *cls._list_params(**kwargs)).execute()


def is_hex(value):
    try:
        _ = int(value, 16)
        return True
    except ValueError:
        return False


def is_valid_sha1(value):
    if len(value) != 40:
        return False
    return is_hex(value)


def is_valid_md5(value):
    if len(value) != 32:
        return False
    return is_hex(value)


def is_valid_sha256(value):
    if len(value) != 64:
        return False
    return is_hex(value)


class Hashable(object):
    SUPPORTED_HASH_TYPES = {
        'sha1': is_valid_sha1,
        'sha256': is_valid_sha256,
        'md5': is_valid_md5,
    }

    def __init__(self, *args, **kwargs):
        # hack to behave as in python 3, signature should be
        # __init__(self, content, *args, hash_value=None, hash_type=None, validate_hash=False, **kwargs)
        hash_value = kwargs.pop('hash_value', None)
        hash_type = kwargs.pop('hash_type', None)
        validate_hash = kwargs.pop('validate_hash', False)
        super(Hashable, self).__init__(*args, **kwargs)

        self._hash = hash_value.strip() if hash_value is not None else None

        if hash_type:
            if hash_type not in self.SUPPORTED_HASH_TYPES:
                raise exceptions.InvalidValueException('Hash type provided is not supported.')
            self._hash_type = hash_type
        else:
            self._hash_type = self.resolve_hash_type()

        if self._hash_type is None:
            raise exceptions.InvalidValueException('Invalid hash provided: {}'.format(self._hash))

        if validate_hash:
            self.validate()

    @property
    def hash(self):
        return self._hash

    @hash.setter
    def hash(self, value):
        self._hash = value.strip() if value is not None else None

    @property
    def hash_type(self):
        return self._hash_type

    def validate(self):
        hash_type = self.resolve_hash_type()
        if self.hash_type != hash_type:
            raise exceptions.InvalidValueException('Detected hash type {}, got type {} for hash {}'
                                                   .format(hash_type, self.hash_type, self.hash))

    def resolve_hash_type(self):
        for hash_type, validator in self.SUPPORTED_HASH_TYPES.items():
            if validator(self._hash):
                return hash_type
        return None

    @property
    def raw(self):
        return unhexlify(self.hash)

    def __eq__(self, other):
        return self.hash == other


def parse_isoformat(date_string):
    """Parses the current date format version """
    if isinstance(date_string, (dt.date, dt.datetime)):
        return date_string
    elif date_string:
        return parser.isoparse(date_string)
    else:
        return None



import logging
import time

import polyswarm_api.core

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from future.utils import string_types

from polyswarm_api import exceptions, resources, settings

logger = logging.getLogger(__name__)


class PolyswarmAPI(object):
    """A synchronous interface to the public and private PolySwarm APIs."""

    def __init__(self, key, uri=None, community=None, timeout=None, verify=True, **kwargs):
        """
        :param key: PolySwarm API key
        :param uri: PolySwarm API URI
        :param community: Community to scan against.
        :param timeout: Maximum time to wait for an http response on every request.
        :param verify: Boolean, whether or not to verify TLS connections.
        :param **kwargs: Keyword args to pass to requests.Session
        """
        key_masked = (key[0:4] if key and len(key) > 16 else '') + '******'
        logger.info('Creating PolyswarmAPI instance: api_key: %s, api_uri: %s, community: %s', key_masked, uri, community)
        self.uri = uri or settings.DEFAULT_GLOBAL_API
        self.community = community or settings.DEFAULT_COMMUNITY
        self.timeout = timeout or settings.DEFAULT_HTTP_TIMEOUT
        self.session = polyswarm_api.core.PolyswarmSession(key, retries=settings.DEFAULT_RETRIES, verify=verify, **kwargs)
        self._engines = None

    @property
    def engines(self):
        if not self._engines:
            self.refresh_engine_cache()

        return self._engines

    def refresh_engine_cache(self):
        """
        Rrefresh the cached engine listing
        """
        engines = list(resources.Engine.list(self).result())
        if not engines:
            raise exceptions.InvalidValueException("Recieved empty engines listing")
        self._engines = engines

    def wait_for(self, scan, timeout=settings.DEFAULT_SCAN_TIMEOUT):
        """
        Wait for a Scan to scan successfully

        :param scan: Scan id to wait for
        :param timeout: Maximum time in seconds to wait before raising a TimeoutException
        :return: The ArtifactInstance resource waited on
        """
        logger.info('Waiting for %s', int(scan))
        start = time.time()
        while True:
            scan_result = self.lookup(scan)
            if scan_result.failed or scan_result.window_closed:
                return scan_result
            elif -1 < timeout < time.time() - start:
                raise exceptions.TimeoutException('Timed out waiting for scan {} to finish. Please try again.'
                                                  .format(scan))
            else:
                time.sleep(settings.POLL_FREQUENCY)

    def exists(self, hash_, hash_type=None):
        """
        Search for the latest scans matching the given hash and hash_type.

        :param hash_: A Hashable object (Artifact, local.LocalArtifact, Hash) or hex-encoded SHA256/SHA1/MD5
        :param hash_type: Hash type of the provided hash_. Will attempt to auto-detect if not explicitly provided.
        :return: A boolean if the instance exists for search.
        """
        logger.info('Exists for hash %s', hash_)
        hash_ = resources.Hash.from_hashable(hash_, hash_type=hash_type)
        result = resources.ArtifactInstance.exists_hash(self, hash_.hash, hash_.hash_type).result()
        if str(result) == '200':
            return True
        else:
            return False

    def search(self, hash_, hash_type=None):
        """
        Search for the latest scans matching the given hash and hash_type.

        :param hash_: A Hashable object (Artifact, local.LocalArtifact, Hash) or hex-encoded SHA256/SHA1/MD5
        :param hash_type: Hash type of the provided hash_. Will attempt to auto-detect if not explicitly provided.
        :return: Generator of ArtifactInstance resources
        """
        logger.info('Searching for hash %s', hash_)
        hash_ = resources.Hash.from_hashable(hash_, hash_type=hash_type)
        return resources.ArtifactInstance.search_hash(self, hash_.hash, hash_.hash_type).result()

    def search_url(self, url):
        """
        Search for the latest scan matching the given url.

        :param url: A url to be searched by exact match
        :return: Generator of ArtifactInstance resources
        """
        logger.info('Searching for url %s', url)
        return resources.ArtifactInstance.search_url(self, url).result()

    def search_scans(self, hash_):
        """
        Search for all scans ever made matching the given sha256.

        :param hash_: A Hashable object (Artifact, local.LocalArtifact, Hash) or hex-encoded SHA256
        :return: Generator of ArtifactInstance resources
        """
        logger.info('Searching for scans %s', hash_)
        hash_ = resources.Hash.from_hashable(hash_, hash_type='sha256')
        return resources.ArtifactInstance.list_scans(self, hash_.hash).result()

    def metadata_mapping(self):
        logger.info('Retrieving the metadata mapping')
        return resources.MetadataMapping.get(self).result()

    def search_by_metadata(self, query, include=None, exclude=None, ips=None, urls=None, domains=None):
        """
        Search artifacts by metadata

        :param query: A query string
        :param include: A list of fields to be included in the result (.* wildcards are accepted)
        :param exclude: A list of fields to be excluded from the result (.* wildcards are accepted)
        :return: Generator of ArtifactInstance resources
        """
        logger.info('Searching for metadata %s', query)
        return resources.Metadata.get(self, query=query, include=include, exclude=exclude, ips=ips, urls=urls, domains=domains).result()

    def iocs_by_hash(self, hash_type, hash_value, hide_known_good=False):
        """
        Retrieve IOCs by artifact hash

        :param hash_type: Hash type of the provided hash_
        :param hash_value: A list of fields to be included in the result (.* wildcards are accepted)
        :return: Generator of IOC resources
        """
        logger.info('Getting IOCs by hash %s:%s', hash_type, hash_value)
        return resources.IOC.iocs_by_hash(self, hash_value, hash_type, hide_known_good=hide_known_good).result()

    def search_by_ioc(self, ip=None, domain=None, ttp=None, imphash=None):
        """
        Search artifacts by IOC (ip, domain, ttp, or imphash)
        
        :param ip: ip address to search by
        :param domain: domain address to search by
        :param ttp: ttp to search by
        :param imphash: ImpHash to search by
        :return: Generator of ArtifactInstance resources
        """
        logger.info('Searching by ioc %s', dict(ip=ip, domain=domain, ttp=ttp, imphash=imphash))
        return resources.IOC.ioc_search(self, ip=ip, domain=domain, ttp=ttp, imphash=imphash).result()

    def check_known_hosts(self, ips=[], domains=[]):
        """
        Check if ip addresses or domains are known.

        :param ips
        :param domains
        :return: Generator of IOC resources
        """
        logger.info('Checking known hosts ips: %s, domains: %s', ips, domains)
        return resources.IOC.check_known_hosts(self, ips, domains).result()

    def add_known_good_host(self, type, source, host):
        """
        Add a known good ip or domain.

        :param type
        :param source
        :param host
        :return: IOC resource
        """
        logger.info('Creating known good ioc %s %s %s', type, host, source)
        return resources.IOC.create_known_good(self, type, host, source).result()

    def update_known_good_host(self, id, type, source, host, good):
        """
        Update a known ip or domain.

        :param type
        :param source
        :param host
        :return: IOC resource
        """
        logger.info('Updating known good ioc %s %s %s %s', id, type, host, source)
        return resources.IOC.update_known_good(self, id, type, host, source, good).result()

    def delete_known_good_host(self, id):
        logger.info('Deleting known good ioc %s', id)
        return resources.IOC.delete_known_good(self, id).result()

    def submit(self, artifact, artifact_type=resources.ArtifactType.FILE, artifact_name=None, scan_config=None):
        """
        Submit artifacts to polyswarm and return UUIDs

        :param artifact: A file-like, path to file, url or LocalArtifact instance
        :param artifact_type: The ArtifactType or strings containing "file" or "url"
        :param artifact_name: An appropriate filename for the Artifact
        :param scan_config: The scan configuration to be used, e.g.: "default", "more-time", "most-time"
        :return: An ArtifactInstance resource
        """
        logger.info('Submitting artifact of type %s', artifact_type)
        artifact_type = resources.ArtifactType.parse(artifact_type)
        # TODO This is a python 2.7 check if artifact is a file-like instance, consider changing
        #  to isinstance(artifact, io.IOBase) when deprecating 2.7 and implementing making LocalHandle
        #  inherit io.IOBase, although this will change the method delegation logic in the resource
        if hasattr(artifact, 'read') and hasattr(artifact.read, '__call__'):
            artifact = resources.LocalArtifact.from_handle(self, artifact, artifact_name=artifact_name or '',
                                                           artifact_type=artifact_type)
        elif isinstance(artifact, string_types):
            if artifact_type == resources.ArtifactType.FILE:
                artifact = resources.LocalArtifact.from_path(self, artifact, artifact_type=artifact_type,
                                                             artifact_name=artifact_name)
            elif artifact_type == resources.ArtifactType.URL:
                artifact = resources.LocalArtifact.from_content(self, artifact, artifact_name=artifact_name or artifact,
                                                                artifact_type=artifact_type)
        if artifact_type == resources.ArtifactType.URL:
            scan_config = scan_config or 'more-time'
        if isinstance(artifact, resources.LocalArtifact):
            instance = resources.ArtifactInstance.create(self,
                                                         artifact_name=artifact.artifact_name,
                                                         artifact_type=artifact.artifact_type.name,
                                                         scan_config=scan_config,
                                                         community=self.community).result()
            instance.upload_file(artifact)
            return resources.ArtifactInstance.update(self, id=instance.id).result()
        else:
            raise exceptions.InvalidValueException('Artifacts should be a path to a file or a LocalArtifact instance')

    def lookup(self, scan):
        """
        Lookup a scan by Scan id.

        :param scan: The Scan UUID to lookup
        :return: An ArtifactInstance resource
        """
        logger.info('Lookup scan %s', int(scan))
        return resources.ArtifactInstance.lookup_uuid(self, scan).result()

    def rescan(self, hash_, hash_type=None, scan_config=None):
        """
        Rescan a file based on and existing hash in the Polyswarm platform

        :param hash_: Hashable object (Artifact, local.LocalArtifact, or Hash) or hex-encoded SHA256/SHA1/MD5
        :param hash_type: Hash type of the provided hash_. Will attempt to auto-detect if not explicitly provided.
        :param scan_config: The scan configuration to be used, e.g.: "default", "more-time", "most-time"
        :return: A ArtifactInstance resources
        """
        logger.info('Rescan hash %s', hash_)
        hash_ = resources.Hash.from_hashable(hash_, hash_type=hash_type)
        return resources.ArtifactInstance.rescan(self, hash_.hash, hash_.hash_type, scan_config=scan_config).result()

    def rescan_id(self, scan, scan_config=None):
        """
        Re-execute a new scan based on an existing scan.

        :param scan: Id of the existing scan
        :param scan_config: The scan configuration to be used, e.g.: "default", "more-time", "most-time"
        :return: A ArtifactInstance resource
        """
        logger.info('Rescan id %s', int(scan))
        return resources.ArtifactInstance.rescan_id(self, scan, scan_config=scan_config).result()

    def _parse_rule(self, rule):
        if isinstance(rule, string_types):
            rule, rule_id = resources.YaraRuleset(dict(yara=rule), api=self), None
        elif isinstance(rule, (resources.YaraRuleset, int)):
            rule, rule_id = None, rule
        else:
            raise exceptions.InvalidValueException('Either yara or rule_id must be provided.')
        return rule, rule_id

    def live_start(self, rule_id):
        """
        Create a new live hunt_id, and replace the currently running YARA rules.

        :param rule_id: Yara ruleset id
        :return: The ruleset with the associated live hunt
        """
        logger.info('Create live hunt for rule id %s', rule_id)
        return resources.LiveYaraRuleset.create(self, rule_id=rule_id).result()

    def live_stop(self, rule_id):
        """
        Stop a live hunt.

        :param rule_id: Yara ruleset id
        :return: The ruleset without an associate live hunt
        """
        logger.info('Delete live hunt for rule id %s', rule_id)
        return resources.LiveYaraRuleset.delete(self, rule_id=rule_id).result()

    def live_feed(self, since=None, rule_name=None, family=None,
                           polyscore_lower=None, polyscore_upper=None):
        """
        Get live hunts feed

        :param since: Fetch results from the last "since" minutes
        :param rule_name: Filter hunt results on the provided rule name (exact match).
        :param family: Filter hunt results based on the family name (exact match).
        :param polyscore_lower: Polyscore lower bound for the hunt results.
        :param polyscore_upper: Polyscore upper bound for the hunt results.
        :return: Generator of HuntResult resources
        """
        return resources.LiveHuntResult.list(
            self, since=since, rule_name=rule_name, family=family,
            polyscore_lower=polyscore_lower, polyscore_upper=polyscore_upper).result()

    def live_feed_delete(self, result_ids):
        """
        Delete live feed results

        :param result_ids: Live Feed Result IDs
        :return: The deleted LiveHuntResult resources
        """
        logger.info('Delete live results: %s', result_ids)
        try:
            return resources.LiveHuntResultList.delete(self, result_ids=result_ids).result()
        except exceptions.NoResultsException:
            return None

    def live_result(self, result_id):
        """
        Get yara ruleset for the live hunt result

        :param result_id: Live result id
        :return: Generator of HuntResult resources
        """
        return resources.LiveHuntResult.get(self, id=result_id).result()

    def historical_create(self, rule=None, ruleset_name=None):
        """
        Run a new historical hunt.

        :param rule: YaraRuleset object or string containing YARA rules to install
        :param ruleset_name: Name of the ruleset.
        :return: The created Hunt resource
        """
        logger.info('Create historical hunt %s', rule)
        rule, rule_id = self._parse_rule(rule)
        return resources.HistoricalHunt.create(self, yara=rule.yara if rule else None, rule_id=rule_id,
                                               ruleset_name=ruleset_name).result()

    def historical_get(self, hunt=None):
        """
        Get a historical hunt.

        :param hunt: Hunt ID
        :return: The Hunt resource
        """
        logger.info('Get historical hunt %s', hunt)
        return resources.HistoricalHunt.get(self, id=hunt).result()

    def historical_update(self, hunt):
        """
        Cancel a historical hunt
        :param hunt: The historical hunt id
        :return: The deleted HistoricalHunt resource
        """
        logger.info('Deleting historical hunt %s', hunt)
        return resources.HistoricalHunt.update(self, id=hunt).result()

    def historical_delete(self, hunt):
        """
        Delete a historical hunts.

        :param hunt: Hunt ID
        :return: The deleted Hunt resource
        """
        logger.info('Delete historical hunt %s', hunt)
        return resources.HistoricalHunt.delete(self, id=hunt).result()

    def historical_list(self, since=None):
        """
        List all historical hunts

        :return: Generator of Hunt resources
        """
        logger.info('List historical hunts since: %s', since)
        return resources.HistoricalHunt.list(self, since=since).result()

    def historical_result(self, result_id):
        """
        Get historical hunt result

        :param result_id: Historical result id
        :return: HistoricalHuntResult resource
        """
        return resources.HistoricalHuntResult.get(self, id=result_id).result()

    def historical_results(self, hunt=None, rule_name=None, family=None,
                           polyscore_lower=None, polyscore_upper=None):
        """
        Get results from a historical hunt

        :param hunt: ID of the hunt (None if latest hunt results are desired)
        :param rule_name: Filter hunt results on the provided rule name (exact match).
        :param family: Filter hunt results based on the family name (exact match).
        :param polyscore_lower: Polyscore lower bound for the hunt results.
        :param polyscore_upper: Polyscore upper bound for the hunt results.
        :return: Generator of HuntResult resources
        """
        logger.info('List historical results for hunt: %s', hunt)
        return resources.HistoricalHuntResultList.get(
            self, id=hunt, rule_name=rule_name, family=family,
            polyscore_lower=polyscore_lower, polyscore_upper=polyscore_upper).result()

    def historical_results_delete(self, result_ids):
        """
        Delete historical scan results

        :param result_ids: Historical Hunt Result IDs
        :return: The deleted HuntResult resources
        """
        logger.info('Delete historical results: %s', result_ids)
        return resources.HistoricalHuntResultList.delete(self, result_ids=result_ids).result()

    def historical_delete_list(self, historical_ids):
        """
        Delete a historical hunts.

        :param historical_ids: Historical Hunt IDs
        :return: The deleted Hunt resource
        """
        logger.info('Delete historical hunts %s', historical_ids)
        return resources.HistoricalHuntList.delete(self, historical_ids=historical_ids).result()

    def ruleset_create(self, name, rules, description=None):
        """
        Create a Yara Ruleset from the provided rules with the given name in the polyswarm platform.
        :param name: Name of the ruleset
        :param rules: Yara rules as a string
        :param description: Description of the ruleset
        :return: A YaraRuleset resource
        """
        logger.info('Create ruleset %s: %s', name, rules)
        rules = resources.YaraRuleset(dict(name=name, description=description, yara=rules), api=self)
        return resources.YaraRuleset.create(self, yara=rules.yara, name=rules.name, description=rules.description).result()

    def ruleset_get(self, ruleset_id=None):
        """
        Retrieve a YaraRuleset from the polyswarm platform by its Id.
        :param ruleset_id: Id of the ruleset
        :return: A YaraRuleset resource
        """
        logger.info('Get ruleset %s', ruleset_id)
        return resources.YaraRuleset.get(self, id=ruleset_id).result()

    def ruleset_update(self, ruleset_id, name=None, rules=None, description=None):
        """
        Update an existing YaraRuleset in the polyswarm platform by its Id.
        :param ruleset_id: Id of the ruleset
        :param name: New name of the ruleset
        :param rules: New yara rules as a string
        :param description: New description of the ruleset
        :return: The updated YaraRuleset resource
        """
        logger.info('Update ruleset %s', ruleset_id)
        return resources.YaraRuleset.update(self, id=ruleset_id, name=name, yara=rules, description=description).result()

    def ruleset_delete(self, ruleset_id):
        """
        Delete a YaraRuleset from the polyswarm platform by its Id.
        :param ruleset_id: Id of the ruleset
        :return: A YaraRuleset resource
        """
        logger.info('Delete ruleset %s', ruleset_id)
        return resources.YaraRuleset.delete(self, id=ruleset_id).result()

    def ruleset_list(self):
        """
        List all YaraRulesets for the current account.
        :return: A generator of YaraRuleset resources
        """
        logger.info('List rulesets')
        return resources.YaraRuleset.list(self).result()

    def tag_link_get(self, sha256):
        """
        Fetch the Tags and Families associated with the given sha256.

        :param sha256: The sha256 of the artifact.
        :return: A TagLink resource
        """
        logger.info('Get tag link %s', sha256)
        return resources.TagLink.get(self, hash=sha256).result()

    def tag_link_update(self, sha256, tags=None, families=None, emerging=None, remove=False):
        """
        Update a TagLink with the given type or value by its id.
        :param sha256: The sha256 of the artifact.
        :param tags: A list of tags to be added or removed.
        :param families: A list of families to be added or removed.
        :param remove: A flag indicating if we should remove the provided tags/families.
        :return: A TagLink resource
        """
        logger.info('Update tag link %s', sha256)
        return resources.TagLink.update(self, hash=sha256, tags=tags, families=families,
                                        emerging=emerging, remove=remove).result()

    def tag_link_list(self, tags=None, families=None, or_tags=None, or_families=None):
        """
        Fetch all existing TagLinks for the provided tags.
        :param tags: A list of tags that must be associated with the TagLinks listed.
        :param families: A list of families that must be associated with the TagLinks listed.
        :param or_tags: A list of tags where the TagLinks must be associated with at least one.
        :param or_families: A list of families where the TagLinks must be associated with at least one.
        :return: A TagLink resource
        """
        logger.info('List tag links')
        return resources.TagLink.list(self, tags=tags, families=families,
                                      or_tags=or_tags, or_families=or_families).result()

    def tag_create(self, name):
        """
        Create a Tag.
        :param name: The tag we want to create.
        :return: A Tag resource
        """
        logger.info('Create tag %s', name)
        return resources.Tag.create(self, name=name).result()

    def tag_get(self, name):
        """
        Fetch a Tag.
        :param name: The tag we want to fetch.
        :return: A Tag resource
        """
        logger.info('Get tag %s', name)
        return resources.Tag.get(self, name=name).result()

    def tag_delete(self, name):
        """
        Delete a Tag.
        :param name: The tag we want to delete.
        :return: A Tag resource
        """
        logger.info('Delete tag %s', name)
        return resources.Tag.delete(self, name=name).result()

    def tag_list(self):
        """
        Fetch all existing Tags.
        :return: A generator of Tag resources
        """
        logger.info('List tags')
        return resources.Tag.list(self).result()

    def family_create(self, name):
        """
        Create a Family.
        :param name: The family name.
        :return: A MalwareFamily resource
        """
        logger.info('Creating family %s', name)
        return resources.MalwareFamily.create(self, name=name).result()

    def family_get(self, name):
        """
        Fetch a Family.
        :param name: The family name.
        :return: A MalwareFamily resource
        """
        logger.info('Getting family %s', name)
        return resources.MalwareFamily.get(self, name=name).result()

    def family_delete(self, name):
        """
        Delete a Family.
        :param name: The family name.
        :return: A MalwareFamily resource
        """
        logger.info('Deleting family %s', name)
        return resources.MalwareFamily.delete(self, name=name).result()

    def family_update(self, family_name, emerging=True):
        """
        Update the Family emerging status.
        :param family_name: The family name.
        :param emerging: A flag indicating if the family should be marked as emerging at this point in time.
        :return: A MalwareFamily resource
        """
        logger.info('Updating family %s', family_name)
        return resources.MalwareFamily.update(self, name=family_name, emerging=emerging).result()

    def family_list(self):
        """
        Fetch all existing Families
        :return: A generator of MalwareFamily resources
        """
        logger.info('Listing families')
        return resources.MalwareFamily.list(self).result()

    def assertions_create(self, engine_id, date_start, date_end):
        logger.info('Create assertions %s %s %s', engine_id, date_start, date_end)
        return resources.AssertionsJob.create(self,
                                              engine_id=engine_id,
                                              date_start=date_start,
                                              date_end=date_end).result()

    def assertions_get(self, assertions_id):
        logger.info('Get assertions %s', assertions_id)
        return resources.AssertionsJob.get(self, id=assertions_id).result()

    def assertions_delete(self, assertions_id):
        logger.info('Delete assertions %s', assertions_id)
        return resources.AssertionsJob.delete(self, id=assertions_id).result()

    def assertions_list(self, engine_id):
        logger.info('Get all assertions bundles for the engine %s', engine_id)
        return resources.AssertionsJob.list(self, engine_id=engine_id).result()

    def votes_create(self, engine_id, date_start, date_end):
        logger.info('Create votes %s %s %s', engine_id, date_start, date_end)
        return resources.VotesJob.create(self,
                                         engine_id=engine_id,
                                         date_start=date_start,
                                         date_end=date_end).result()

    def votes_get(self, votes_id):
        logger.info('Get votes %s', votes_id)
        return resources.VotesJob.get(self, id=votes_id).result()

    def votes_delete(self, votes_id):
        logger.info('Delete votes %s', votes_id)
        return resources.VotesJob.delete(self, id=votes_id).result()

    def votes_list(self, engine_id):
        logger.info('Get all votes bundles for the engine %s', engine_id)
        return resources.VotesJob.list(self, engine_id=engine_id).result()

    def download(self, out_dir, hash_, hash_type=None):
        """
        Grab the data of artifact identified by hash, and write the data to a file in the provided directory
        under a file named after the hash_.
        :param out_dir: Destination directory to download the file.
        :param hash_: The hash we should use to lookup the artifact to download.
        :param hash_type: Hash type of the provided hash_. Will attempt to auto-detect if not explicitly provided.
        :return: A LocalArtifact resource
        """
        logger.info('Downloading %s into %s', hash_, out_dir)
        hash_ = resources.Hash.from_hashable(hash_, hash_type=hash_type)
        artifact = resources.LocalArtifact.download(self, hash_.hash, hash_.hash_type, folder=out_dir).result()
        artifact.handle.close()

        return artifact

    def sandbox(self, sha256_hash, force=False):
        """
        Send a scanned artifact by hash to the sandboxing system.
        :param sha256_hash:
        :param force: 
        """
        logger.info('Sandboxing %s force=%s', sha256_hash, force)
        return resources.SandboxResult.sandbox(self, sha256_hash, force=force)

    def download_archive(self, out_dir, s3_path):
        """
        Grab the data in the s3 path provided in the stream() method, and write the contents
        in the provided directory.
        :param out_dir: Destination directory to download the file.
        :param s3_path: Target S3 object to download.
        :return: A LocalArtifact resource
        """
        logger.info('Downloading %s into %s', s3_path, out_dir)
        artifact = resources.LocalArtifact.download_archive(self, s3_path, folder=out_dir).result()
        artifact.handle.close()

        return artifact

    def download_to_handle(self, hash_, fh, hash_type=None):
        """
        Grab the data of artifact identified by hash, and write the data to a file handle
        :param hash_: The hash we should use to lookup the artifact to download.
        :param fh: A file-like object which we are going to write the contents of the artifact to.
        :param hash_type: Hash type of the provided hash_. Will attempt to auto-detect if not explicitly provided.
        :return: A LocalHandle resource
        """
        logger.info('Downloading %s into handle', hash_)
        hash_ = resources.Hash.from_hashable(hash_, hash_type=hash_type)
        return resources.LocalArtifact.download(self, hash_.hash, hash_.hash_type, handle=fh).result()

    def stream(self, since=settings.MAX_SINCE_TIME_STREAM):
        """
        Access the stream of artifacts (ask info@polyswarm.io about access)

        :param since: Fetch results from the last "since" minutes (up to 2 days)
        :return: Generator of ArtifactArchive resources
        """
        logger.info('Streaming since %s', since)
        return resources.ArtifactArchive.get(self, since=since).result()

    def rerun_metadata(self, hashes, analyses=None, skip_es=None):
        logger.info('Rerunning metadata for hashes %s', hashes)
        return resources.ArtifactInstance.metadata_rerun(self, hashes, analyses=analyses, skip_es=skip_es).result()

    def tool_metadata_create(self, sha256, tool, tool_metadata):
        logger.info('Create tool metadata %s %s %s', sha256, tool, tool_metadata)
        return resources.ToolMetadata.create(self, sha256=sha256, tool=tool, tool_metadata=tool_metadata).result()

    def tool_metadata_list(self, sha256):
        logger.info('List tool metadata')
        return resources.ToolMetadata.list(self, sha256=sha256).result()

    def __repr__(self):
        clsname = '{0.__module__}.{0.__name__}'.format(self.__class__)
        attrs = 'uri={0.uri!r}, community={0.community!r}, timeout={0.timeout!r}'.format(self)
        return '<{}({}) at 0x{:x}>'.format(clsname, attrs, id(self))



import logging
import os
import io
import functools
import requests
from enum import Enum
from hashlib import sha256 as _sha256, sha1 as _sha1, md5 as _md5

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from future.utils import raise_from, string_types

# Windows might rase an OSError instead of an ImportError like this
# OSError: [WinError 193] %1 is not a valid Win32 application
try:
    import yara
except (ImportError, OSError):
    yara = None

from polyswarm_api import exceptions, core, settings

logger = logging.getLogger(__name__)

#####################################################################
# Resources returned by the API
#####################################################################


class Engine(core.BaseJsonResource):
    RESOURCE_ENDPOINT = '/microengines'

    def __init__(self, content, api=None):
        super(Engine, self).__init__(content=content, api=api)
        self.id = str(content['id'])
        self.name = content['name']

        try:
            self.address = content['address'].lower()
        except:
            self.address = None

        account_number = content.get('accountNumber')
        self.account_number = str(account_number) if account_number else None

        self.engine_type = content.get('engineType', 'microengine')
        self.is_microengine = self.engine_type == 'microengine'
        self.is_arbiter = self.engine_type == 'arbiter'

        self.status = content.get('status', 'disabled')
        self.verified = self.status == 'verified'

        # These fields can be `null`; don't replace w/ default value in `get()`
        self.artifact_types = set(content.get('artifactTypes') or [])
        self.tags = set(content.get('tags') or [])
        self.communities = set(content.get('communities') or [])
        self.mimetypes = set(content.get('mimeTypes') or [])

        self.created_at = core.parse_isoformat(content.get('createdAt'))
        self.modified_at = core.parse_isoformat(content.get('modifiedAt'))
        self.archived_at = core.parse_isoformat(content.get('archivedAt'))

    @classmethod
    def _list_headers(cls, api):
        return {'Authorization': None}

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.id == other.id if isinstance(other, Engine) else False

    def __repr__(self):
        return '{}(name={}, address={}, status={}, engine_type={})'.format(
            self.__class__.__name__,
            self.name,
            self.address,
            self.status,
            self.engine_type,
        )


class ToolMetadata(core.BaseJsonResource):
    RESOURCE_ENDPOINT = '/artifact/metadata'


class MetadataMapping(core.BaseJsonResource):
    RESOURCE_ENDPOINT = '/search/metadata/mappings'


class Metadata(core.BaseJsonResource):
    RESOURCE_ENDPOINT = '/search/metadata/query'
    KNOWN_KEYS = {'artifact', 'exiftool', 'hash', 'lief', 'pefile', 'scan', 'strings'}

    def __init__(self, content, api=None):
        super(Metadata, self).__init__(content=content, api=api)
        self.created = core.parse_isoformat(self.artifact.get('created'))

        self.id = self._get('artifact.id')
        self.sha1 = self._get('artifact.sha1')
        self.sha256 = self._get('artifact.sha256')
        self.md5 = self._get('artifact.md5')

        self.ssdeep = self._get('hash.ssdeep')
        self.tlsh = self._get('hash.tlsh')

        self.first_seen = core.parse_isoformat(self._get('scan.first_scan.created'))
        self.last_scanned = core.parse_isoformat(self._get('scan.latest_scan.created'))
        self.mimetype = self._get('scan.mimetype.mime')
        self.extended_mimetype = self._get('scan.mimetype.extended')
        self.malicious = self._get('scan.detections.malicious')
        self.benign = self._get('scan.detections.benign')
        self.total_detections = self._get('scan.detections.total')
        self.filenames = self._get('scan.filename')

        self.domains = self._get('strings.domains')
        self.ipv4 = self._get('strings.ipv4')
        self.ipv6 = self._get('strings.ipv6')
        self.urls = self._get('strings.urls')

    def __contains__(self, item):
        return item in self.json

    def __getattr__(self, name):
        try:
            return self.json[name]
        except KeyError:
            if name in Metadata.KNOWN_KEYS:
                return {}
            raise AttributeError()

    @classmethod
    def _get_params(cls, **kwargs):
        params = []
        include = kwargs.pop('include', ()) or ()
        exclude = kwargs.pop('exclude', ()) or ()
        ips = kwargs.pop('ips', ()) or ()
        urls = kwargs.pop('urls', ()) or ()
        domains = kwargs.pop('domains', ()) or ()
        params.extend(('include', v) for v in include)
        params.extend(('exclude', v) for v in exclude)
        params.extend(('ips', v) for v in ips)
        params.extend(('urls', v) for v in urls)
        params.extend(('domains', v) for v in domains)
        super_params, json_params = super(Metadata, cls)._get_params(**kwargs)
        params.extend(super_params.items())
        return params, json_params


class IOC(core.BaseJsonResource):
    RESOURCE_ENDPOINT = '/ioc'

    @classmethod
    def iocs_by_hash(cls, api, hash_value, hash_type, hide_known_good=False):
        return core.PolyswarmRequest(
            api,
            {
                'method': 'GET',
                'url': '{}/ioc/{}/{}'.format(api.uri, hash_type, hash_value),
                'params': {
                    'hide_known_good': hide_known_good,
                },
            },
            result_parser=cls,
        ).execute()

    @classmethod
    def ioc_search(cls, api, ip=None, domain=None, ttp=None, imphash=None):
        params = dict()
        if ip is not None:
            params['ip'] = ip
        if domain is not None:
            params['domain'] = domain
        if ttp is not None:
            params['ttp'] = ttp
        if imphash is not None:
            params['imphash'] = imphash
        return core.PolyswarmRequest(
            api,
            {
                'method': 'GET',
                'url': '{}/ioc/search'.format(api.uri),
                'params': params
            },
            result_parser=cls,
        ).execute()

    @classmethod
    def check_known_hosts(cls, api, ips, domains):
        return core.PolyswarmRequest(
            api,
            {
                'method': 'GET',
                'url': '{}/ioc/known'.format(api.uri),
                'params': {
                    'ip': ips,
                    'domain': domains
                }
            },
            result_parser=cls,
        ).execute()

    @classmethod
    def create_known_good(cls, api, type, host, source):
        return core.PolyswarmRequest(
            api,
            {
                'method': 'POST',
                'url': '{}/ioc/known'.format(api.uri),
                'json': {
                    'type': type,
                    'host': host,
                    'source': source,
                    'good': True
                }
            },
            result_parser=cls,
        ).execute()

    @classmethod
    def update_known_good(cls, api, id, type, host, source, good):
        return core.PolyswarmRequest(
            api,
            {
                'method': 'PUT',
                'url': '{}/ioc/known'.format(api.uri),
                'json': {
                    'id': id,
                    'type': type,
                    'host': host,
                    'source': source,
                    'good': good
                }
            },
            result_parser=cls,
        ).execute()

    @classmethod
    def delete_known_good(cls, api, id):
        return core.PolyswarmRequest(
            api,
            {
                'method': 'DELETE',
                'url': '{}/ioc/known'.format(api.uri),
                'params': {
                    'id': id
                }
            },
            result_parser=cls,
        ).execute()
        
class ArtifactInstance(core.BaseJsonResource, core.Hashable):
    RESOURCE_ENDPOINT = '/instance'

    def __init__(self, content, api=None):
        super(ArtifactInstance, self).__init__(content=content, api=api,
                                               hash_value=content['sha256'], hash_type='sha256')
        # Artifact fields
        self.sha256 = content['sha256']
        self.artifact_id = content.get('artifact_id')
        self.md5 = content['md5']
        self.sha1 = content['sha1']
        self.mimetype = content['mimetype']
        self.size = content['size']
        self.extended_type = content['extended_type']
        self.first_seen = core.parse_isoformat(content['first_seen'])
        self.upload_url = content['upload_url']
        # Deprecated
        self.last_seen = core.parse_isoformat(content.get('last_seen'))
        self.last_scanned = core.parse_isoformat(content.get('last_scanned'))
        metadata_json = content.get('metadata') or []
        metadata = {metadata['tool']: metadata['tool_metadata'] for metadata in metadata_json}
        self.metadata = Metadata(metadata, api)

        # ArtifactInstance fields
        self.id = content.get('id')
        self.assertions = [Assertion(a, api=api, scanfile=self) for a in content.get('assertions', [])]
        self.country = content.get('country')
        self.community = content.get('community')
        self.created = core.parse_isoformat(content.get('created'))
        self.failed = content.get('failed')
        self.filename = content.get('filename')
        self.result = content.get('result')
        self.type = content.get('type')
        self.votes = [Vote(v, api=api, scanfile=self) for v in content.get('votes', [])]
        self.window_closed = content.get('window_closed')
        self.polyscore = float(content['polyscore']) if content.get('polyscore') is not None else None
        self.permalink = settings.DEFAULT_PERMALINK_BASE + '/' + str(self.hash)

        self._malicious_assertions = None
        self._benign_assertions = None
        self._valid_assertions = None

    def upload_file(self, artifact, attempts=3, **kwargs):
        if not self.upload_url:
            raise exceptions.InvalidValueException('upload_url must be set to upload a file')
        if not artifact:
            raise exceptions.InvalidValueException('A LocalArtifact must be provided in order to upload')
        r = None
        while attempts > 0 and not r:
            attempts -= 1
            artifact.seek(0, io.SEEK_END)
            length = artifact.tell()
            artifact.seek(0)
            # https://github.com/psf/requests/issues/4215#issuecomment-319521235
            # We have to manually handle the case when the file is empty
            # in a way that requests won't set Transfer-Encoding: chunked
            if not length:
                artifact = ''
            r = requests.put(self.upload_url, data=artifact, **kwargs)
            r.raise_for_status()
        return r

    @classmethod
    def exists_hash(cls, api, hash_value, hash_type):
        return core.PolyswarmRequest(
            api,
            {
                'method': 'HEAD',
                'url': '{}/search/hash/{}'.format(api.uri, hash_type),
                'params': {
                    'hash': hash_value,
                },
            },
        ).execute()

    @classmethod
    def search_hash(cls, api, hash_value, hash_type):
        return core.PolyswarmRequest(
            api,
            {
                'method': 'GET',
                'url': '{}/search/hash/{}'.format(api.uri, hash_type),
                'params': {
                    'hash': hash_value,
                },
            },
            result_parser=cls,
        ).execute()

    @classmethod
    def search_url(cls, api, url):
        return core.PolyswarmRequest(
            api,
            {
                'method': 'GET',
                'url': '{}/search/url'.format(api.uri),
                'params': {
                    'url': url,
                },
            },
            result_parser=cls,
        ).execute()

    @classmethod
    def list_scans(cls, api, hash_value):
        return core.PolyswarmRequest(
            api,
            {
                'method': 'GET',
                'url': '{}/search/instances'.format(api.uri),
                'params': {
                    'hash': hash_value,
                },
            },
            result_parser=cls,
        ).execute()

    @classmethod
    def submit(cls, api, artifact, artifact_name, artifact_type, scan_config=None):
        parameters = {
            'method': 'POST',
            'url': '{}/consumer/submission/{}'.format(api.uri, api.community),
            'files': {
                'file': (artifact_name, artifact),
            },
            # very oddly, when included in files parameter this errors out
            'data': {
                'artifact-type': artifact_type,
            }
        }
        if scan_config:
            parameters['data']['scan-config'] = scan_config
        return core.PolyswarmRequest(
            api,
            parameters,
            result_parser=cls,
        ).execute()

    @classmethod
    def rescan(cls, api, hash_value, hash_type, scan_config=None):
        parameters = {
            'method': 'POST',
            'url': '{}/consumer/submission/{}/rescan/{}/{}'.format(api.uri, api.community, hash_type, hash_value),
        }
        if scan_config:
            parameters.setdefault('data', {})['scan-config'] = scan_config
        return core.PolyswarmRequest(
            api,
            parameters,
            result_parser=cls,
        ).execute()

    @classmethod
    def rescan_id(cls, api, submission_id, scan_config=None):
        parameters = {
            'method': 'POST',
            'url': '{}/consumer/submission/{}/rescan/{}'.format(api.uri, api.community, int(submission_id)),
        }
        if scan_config:
            parameters.setdefault('data', {})['scan-config'] = scan_config
        return core.PolyswarmRequest(
            api,
            parameters,
            result_parser=cls,
        ).execute()

    @classmethod
    def lookup_uuid(cls, api, submission_id):
        return core.PolyswarmRequest(
            api,
            {
                'method': 'GET',
                'url': '{}/consumer/submission/{}/{}'.format(api.uri, api.community, int(submission_id)),
            },
            result_parser=cls,
        ).execute()

    @classmethod
    def metadata_rerun(cls, api, hashes, analyses=None, skip_es=None):
        parameters = {
            'method': 'POST',
            'url': '{}/consumer/metadata'.format(api.uri),
            'json': {'hashes': hashes},
        }
        if analyses:
            parameters['json']['analyses'] = analyses
        if skip_es:
            parameters['json']['skip_es'] = skip_es
        return core.PolyswarmRequest(
            api,
            parameters,
            result_parser=cls,
        ).execute()

    def __str__(self):
        return "ArtifactInstance-<%s>" % self.hash

    @property
    def malicious_assertions(self):
        if not self._malicious_assertions:
            self._malicious_assertions = [a for a in self.assertions if a.mask and a.verdict]
        return self._malicious_assertions

    @property
    def benign_assertions(self):
        if not self._benign_assertions:
            self._benign_assertions = [a for a in self.assertions if a.mask and not a.verdict]
        return self._benign_assertions

    @property
    def valid_assertions(self):
        if not self._valid_assertions:
            self._valid_assertions = [a for a in self.assertions if a.mask]
        return self._valid_assertions


class ArtifactArchive(core.BaseJsonResource):
    RESOURCE_ENDPOINT = '/consumer/download/stream'

    def __init__(self, content, api=None):
        super(ArtifactArchive, self).__init__(content=content, api=api)
        self.id = content['id']
        self.community = content['community']
        self.created = core.parse_isoformat(content['created'])
        self.uri = content['uri']


class AssertionsJob(core.BaseJsonResource):
    RESOURCE_ENDPOINT = '/consumer/assertions-job'

    def __init__(self, content, api=None):
        super(AssertionsJob, self).__init__(content=content, api=api)
        self.id = content['id']
        self.engine_id = content['engine_id']
        self.created = core.parse_isoformat(content['created'])
        self.date_start = core.parse_isoformat(content['date_start'])
        self.date_end = core.parse_isoformat(content['date_end'])
        self.storage_path = content['storage_path']
        self.true_positive = content['true_positive']
        self.true_negative = content['true_negative']
        self.false_positive = content['false_positive']
        self.false_negative = content['false_negative']
        self.suspicious = content['suspicious']
        self.unknown = content['unknown']
        self.total = content['total']


class VotesJob(core.BaseJsonResource):
    RESOURCE_ENDPOINT = '/consumer/votes-job'

    def __init__(self, content, api=None):
        super(VotesJob, self).__init__(content=content, api=api)
        self.id = content['id']
        self.engine_id = content['engine_id']
        self.created = core.parse_isoformat(content['created'])
        self.date_start = core.parse_isoformat(content['date_start'])
        self.date_end = core.parse_isoformat(content['date_end'])
        self.storage_path = content['storage_path']
        self.true_positive = content['true_positive']
        self.true_negative = content['true_negative']
        self.false_positive = content['false_positive']
        self.false_negative = content['false_negative']
        self.suspicious = content['suspicious']
        self.unknown = content['unknown']
        self.total = content['total']


def _read_chunks(file_handle):
    while True:
        data = file_handle.read(settings.FILE_CHUNK_SIZE)
        if not data:
            break
        yield data


def all_hashes(file_handle, algorithms=(_sha256, _sha1, _md5)):
    hashers = [alg() for alg in algorithms]
    for data in _read_chunks(file_handle):
        [h.update(data) for h in hashers]
    return [Hash(h.hexdigest()) for h in hashers]


class LocalArtifact(core.BaseResource, core.Hashable):
    """ Artifact for which we have local content """
    def __init__(self, response, api=None, handle=None, folder=None,
                 artifact_name=None, artifact_type=None, analyze=False, **kwargs):
        """
        A representation of an artifact we have locally

        :param artifact_name: Name of the artifact
        :param artifact_type: Type of artifact
        :param api: PolyswarmAPI instance
        :param analyze: Boolean, if True will run analyses on artifact on startup (Note: this may still run later if False)
        """
        # check if we have a destination to store the file
        # raise an error if we don't have exacltly one
        if folder and handle:
            raise exceptions.InvalidValueException('Only one of path or handle should be defined.')
        if not (folder or handle):
            raise exceptions.InvalidValueException('At least one of path or handle must be defined.')

        # initialize super classes and default values
        super(LocalArtifact, self).__init__(response, api=api, hash_type='sha256')
        self.sha256 = None
        self.sha1 = None
        self.md5 = None
        self.analyzed = False
        self.artifact_type = artifact_type or ArtifactType.FILE

        # resolve the file name
        if artifact_name:
            # prioritize explicitly provided name
            self.artifact_name = artifact_name
        else:
            if response:
                # respect content-disposition if there is a response
                filename = response.headers.get('content-disposition', '').partition('filename=')[2]
                if filename:
                    self.artifact_name = filename
                elif os.path.basename(getattr(handle, 'name', '')):
                    self.artifact_name = os.path.basename(getattr(handle, 'name', ''))
                else:
                    self.artifact_name = os.path.basename(urlparse(response.url).path)
            elif os.path.basename(getattr(handle, 'name', '')):
                # if there is no response and no artifact_name, try to get from the handle
                self.artifact_name = os.path.basename(getattr(handle, 'name', ''))

        # resolve the handle to be used
        # only one of handle or folder can be provided (we checked for this above)
        # if one was explicitly provided, use it
        # if we have a folder, use a file named after file_name in that folder
        # otherwise use an in-memory handle
        remove_on_error = False
        try:
            if folder:
                # TODO: this should be replaced with os.makedirs(path, exist_ok=True)
                #  once we drop support to python 2.7
                if not os.path.exists(folder):
                    try:
                        os.makedirs(folder)
                    except FileExistsError:
                        pass
                remove_on_error = True
                self.handle = open(os.path.join(folder, self.artifact_name), mode='wb+', **kwargs)
            else:
                self.handle = handle or io.BytesIO()

            if response:
                # process the content in the response if available, write to handle
                for chunk in response.iter_content(settings.DOWNLOAD_CHUNK_SIZE):
                    self.handle.write(chunk)
                    if hasattr(self.handle, 'flush'):
                        self.handle.flush()
            if analyze:
                # analyze the artifact in case it is needed
                self.analyze_artifact()
        except Exception:
            try:
                if remove_on_error and self.handle:
                    # make sure we cleanup the handle
                    # if an exception happened and this is a file we created
                    self.handle.close()
                    os.remove(self.handle.name)
            except Exception:
                logger.exception('Failed to cleanup the target file.')
            raise

    @classmethod
    def download(cls, api, hash_value, hash_type, handle=None, folder=None, artifact_name=None):
        return core.PolyswarmRequest(
            api,
            {
                'method': 'GET',
                'url': '{}/consumer/download/{}/{}'.format(api.uri, hash_type, hash_value),
                'stream': True,
            },
            result_parser=cls,
            handle=handle,
            folder=folder,
            artifact_name=artifact_name,
        ).execute()

    @classmethod
    def download_archive(cls, api, u, handle=None, folder=None, artifact_name=None):
        """ This method is special, in that it is simply for downloading from S3 """
        return core.PolyswarmRequest(
            api,
            {
                'method': 'GET',
                'url': u,
                'stream': True,
                'headers': {'Authorization': None}
            },
            result_parser=cls,
            handle=handle,
            folder=folder,
            artifact_name=artifact_name,
        ).execute()

    # Inspired by
    # https://github.com/python/cpython/blob/29500737d45cbca9604d9ce845fb2acc3f531401/Lib/tempfile.py#L461
    def __getattr__(self, name):
        # Attribute lookups are delegated to the underlying file
        # and cached for non-numeric results
        # (i.e. methods are cached, closed and friends are not)
        a = getattr(self.handle, name)

        if hasattr(a, '__call__'):
            func = a

            @functools.wraps(func)
            def func_wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            a = func_wrapper
        if not isinstance(a, int):
            setattr(self, name, a)
        return a

    # https://github.com/python/cpython/blob/29500737d45cbca9604d9ce845fb2acc3f531401/Lib/tempfile.py#L499
    # iter() doesn't use __getattr__ to find the __iter__ method
    def __iter__(self):
        for line in self.handle:
            yield line

    @classmethod
    def from_handle(cls, api, handle, artifact_type=None, analyze=False, artifact_name=None, **kwargs):
        # create the LocalHandle with the given handle and don't write anything to it
        return cls(b'', handle=handle, artifact_name=artifact_name,
                   artifact_type=artifact_type, analyze=analyze, api=api)

    @classmethod
    def from_path(cls, api, path, artifact_type=None, analyze=False, artifact_name=None, **kwargs):
        if not isinstance(path, string_types):
            raise exceptions.InvalidValueException('Path should be a string')
        artifact_name = artifact_name or os.path.basename(path)
        handle = open(path, mode='rb', **kwargs)
        # create the LocalHandle with the given handle and don't write anything to it
        return cls(b'', handle=handle, artifact_name=artifact_name,
                   artifact_type=artifact_type, analyze=analyze, api=api)

    @classmethod
    def from_content(cls, api, content, artifact_name=None, artifact_type=None, analyze=False):
        if isinstance(content, string_types):
            content = content.encode("utf8")
        handle = io.BytesIO(content)
        # create the LocalHandle with the given handle and don't write anything to it
        return cls(b'', handle=handle, artifact_name=artifact_name,
                   artifact_type=artifact_type, analyze=analyze, api=api)

    @property
    def hash(self):
        self.analyze_artifact()
        return super(LocalArtifact, self).hash

    def analyze_artifact(self, force=False):
        if not self.analyzed or force:
            self.handle.seek(0)
            self._calc_hashes(self.handle)
            self.handle.seek(0)
            self._run_analyzers(self.handle)
            self.analyzed = True
            # define the hash value only when analyzed
            self._hash = self.sha256

    def _calc_hashes(self, fh):
        self.sha256, self.sha1, self.md5 = all_hashes(fh)

    def _run_analyzers(self, fh):
        # TODO implement custom analyzer support, so users can implement plugins here.
        return {}

    def __str__(self):
        return "Artifact <%s>" % self.hash


class YaraRuleset(core.BaseJsonResource):
    RESOURCE_ENDPOINT = '/hunt/rule'

    def __init__(self, content, api=None):
        super(YaraRuleset, self).__init__(content, api=api)
        self.id = content.get('id')
        self.livescan_id = content.get('livescan_id')
        self.livescan_created = content.get('livescan_created')
        self.name = content.get('name')
        self.description = content.get('description')
        self.created = core.parse_isoformat(content.get('created'))
        self.modified = core.parse_isoformat(content.get('modified'))
        self.deleted = content.get('deleted')
        self.yara = content.get('yara')


class LiveYaraRuleset(YaraRuleset):
    RESOURCE_ENDPOINT = '/hunt/rule/live'


class LiveHuntResult(core.BaseJsonResource):
    RESOURCE_ENDPOINT = '/hunt/live'

    def __init__(self, content, api=None):
        super(LiveHuntResult, self).__init__(content=content, api=api)
        self.id = content['id']
        self.livescan_id = content['livescan_id']
        self.instance_id = content['instance_id']
        self.created = core.parse_isoformat(content['created'])
        self.sha256 = content['sha256']
        self.rule_name = content['rule_name']
        self.tags = content['tags']
        self.polyscore = content['polyscore']
        self.malware_family = content['malware_family']
        self.detections = content['detections']
        self.yara = content.get('yara')
        self.download_url = content.get('download_url')


class LiveHuntResultList(LiveHuntResult):
    RESOURCE_ENDPOINT = '/hunt/live/list'


class HistoricalHunt(core.BaseJsonResource):
    RESOURCE_ENDPOINT = '/hunt/historical'

    def __init__(self, content, api=None):
        super(HistoricalHunt, self).__init__(content=content, api=api)
        # active only present for live   hunts
        self.id = content['id']
        self.created = core.parse_isoformat(content['created'])
        self.status = content['status']
        self.active = content.get('active')
        self.ruleset_name = content.get('ruleset_name')
        self.yara = content.get('yara')
        self.summary = content.get('summary')
        self.progress = content['progress']
        self.results_csv_uri = content['results_csv_uri']


class HistoricalHuntList(HistoricalHunt):
    RESOURCE_ENDPOINT = '/hunt/historical/list'


class HistoricalHuntResult(core.BaseJsonResource):
    RESOURCE_ENDPOINT = '/hunt/historical/results'

    def __init__(self, content, api=None):
        super(HistoricalHuntResult, self).__init__(content=content, api=api)
        self.id = content['id']
        self.historicalscan_id = content['historicalscan_id']
        self.instance_id = content['instance_id']
        self.sha256 = content['sha256']
        self.created = core.parse_isoformat(content['created'])
        self.rule_name = content['rule_name']
        self.tags = content['tags']
        self.polyscore = content['polyscore']
        self.malware_family = content['malware_family']
        self.detections = content['detections']
        self.download_url = content.get('download_url')


class HistoricalHuntResultList(HistoricalHuntResult):
    RESOURCE_ENDPOINT = '/hunt/historical/results/list'


class TagLink(core.BaseJsonResource):
    RESOURCE_ENDPOINT = '/tags/link'
    RESOURCE_ID_KEY = 'hash'

    def __init__(self, content, api=None):
        super(TagLink, self).__init__(content, api=api)
        self.id = content.get('id')
        self.sha256 = content.get('sha256')
        self.created = core.parse_isoformat(content.get('created'))
        self.updated = core.parse_isoformat(content.get('updated'))
        self.first_seen = core.parse_isoformat(content.get('first_seen'))
        self.tags = content.get('tags')
        self.families = content.get('families')
        self.emerging = core.parse_isoformat(content.get('emerging'))

    @classmethod
    def _list_params(cls, **kwargs):
        params = []
        empty = tuple()
        params.extend(('tag', p) for p in kwargs.get('tags', empty))
        params.extend(('family', p) for p in kwargs.get('families', empty))
        params.extend(('or_tag', p) for p in kwargs.get('or_tags', empty))
        params.extend(('or_family', p) for p in kwargs.get('or_families', empty))
        params.append(('emerging', kwargs.get('emerging', empty)))
        return params, None


class MalwareFamily(core.BaseJsonResource):
    RESOURCE_ENDPOINT = '/tags/family'
    RESOURCE_ID_KEY = 'name'

    def __init__(self, content, api=None):
        super(MalwareFamily, self).__init__(content, api=api)
        self.id = content.get('id')
        self.created = core.parse_isoformat(content.get('created'))
        self.updated = core.parse_isoformat(content.get('updated'))
        self.name = content.get('name')
        self.emerging = core.parse_isoformat(content.get('emerging'))


class Tag(core.BaseJsonResource):
    RESOURCE_ENDPOINT = '/tags/tag'
    RESOURCE_ID_KEY = 'name'

    def __init__(self, content, api=None):
        super(Tag, self).__init__(content, api=api)
        self.id = content.get('id')
        self.created = core.parse_isoformat(content.get('created'))
        self.updated = core.parse_isoformat(content.get('updated'))
        self.name = content.get('name')


#####################################################################
# Nested Resources
#####################################################################


class Assertion(core.BaseJsonResource):
    def __init__(self, content, api=None, scanfile=None):
        super(Assertion, self).__init__(content, api=api)
        self.scanfile = scanfile
        self.author = content['author']
        self.author_name = content['author_name']
        self.engine_name = content['engine'].get('name')
        self.bid = int(content['bid'])
        self.mask = content['mask']
        # deal with metadata being a string instead of null
        self.metadata = content['metadata'] if content['metadata'] else {}
        self.verdict = content['verdict']

    def __str__(self):
        return "Assertion-%s: %s" % (self.engine_name, self.verdict)


class Vote(core.BaseJsonResource):
    def __init__(self, content, api=None, scanfile=None):
        super(Vote, self).__init__(content, api=api)
        self.scanfile = scanfile
        self.arbiter = content['arbiter']
        self.vote = content['vote']

    def __str__(self):
        return "Vote-%s: %s" % (self.arbiter, self.vote)


#####################################################################
# Resources Used as input parameters in PolyswarmAPI
#####################################################################


class ArtifactType(Enum):
    FILE = 0
    URL = 1

    @staticmethod
    def parse(value):
        if isinstance(value, ArtifactType):
            return value
        try:
            return ArtifactType[value.upper()]
        except Exception as e:
            raise raise_from(
                exceptions.InvalidValueException(
                    'Unable to get the artifact type from the provided value {}'.format(value)), e)

    @staticmethod
    def to_string(artifact_type):
        return artifact_type.name.lower()

    def decode_content(self, content):
        if content is None:
            return None

        if self == ArtifactType.URL:
            try:
                return content.decode('utf-8')
            except UnicodeDecodeError:
                raise exceptions.DecodeErrorException('Error decoding URL')
        else:
            return content


class Hash(core.Hashable):
    def __init__(self, hash_, hash_type=None, validate_hash=True):
        super(Hash, self).__init__(hash_value=hash_, hash_type=hash_type, validate_hash=validate_hash)

    @classmethod
    def from_hashable(cls, hash_, hash_type=None):
        """
        Coerce to Hashable object

        :param hash_: Hashable object
        :param hash_type: Hash type
        :param polyswarm: PolyswarmAPI instance
        :return: Hash
        """
        if issubclass(type(hash_), core.Hashable):
            if hash_type and hash_.hash_type != hash_type:
                raise exceptions.InvalidValueException('Detected hash type {}, got {} for hashable {}'
                                                       .format(hash_.hash_type, hash_type, hash_.hash))
            return Hash(hash_.hash, hash_type=hash_type)
        return Hash(hash_, hash_type=hash_type)

    def __hash__(self):
        return hash(self.hash)

    def __str__(self):
        return self.hash

    def __repr__(self):
        return "{}={}".format(self.hash_type, self.hash)


class SandboxResult(core.BaseJsonResource):

    def __init__(self, content, api=None):
        super(SandboxResult, self).__init__(content, api=api)

    @classmethod
    def sandbox(cls, api, sha256, force=False):
        return core.PolyswarmRequest(
            api,
            {
                'method':
                'POST',
                'url':
                '{}/consumer/submission/{}/sandbox/{}'.format(api.uri, api.community, sha256),
                'params': {
                    'force': force,
                },
            },
            result_parser=cls,
        ).execute()



# https://www.python.org/dev/peps/pep-0008/#module-level-dunder-names
__version__ = '3.1.0'
__release_url__ = 'https://api.github.com/repos/polyswarm/polyswarm-api/releases/latest'

from . import api
from . import exceptions
